@@ -1,9 +1,11 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 
+from trytond.transaction import Transaction
 from trytond.backend.table import TableHandlerInterface
 import logging
 import re
+import warnings
 
 __all__ = ['TableHandler']
 
@@ -11,31 +13,32 @@ logger = logging.getLogger(__name__)
 
 
 class TableHandler(TableHandlerInterface):
-    def __init__(self, cursor, model, module_name=None, history=False):
-        super(TableHandler, self).__init__(cursor, model,
+    def __init__(self, model, module_name=None, history=False):
+        super(TableHandler, self).__init__(model,
                 module_name=module_name, history=history)
         self._columns = {}
         self._constraints = []
         self._fk_deltypes = {}
         self._indexes = []
-        self._field2module = {}
         self._model = model
 
+        cursor = Transaction().connection.cursor()
         # Create new table if necessary
-        if not self.table_exist(self.cursor, self.table_name):
+        if not self.table_exist(self.table_name):
             if not self.history:
-                self.cursor.execute('CREATE TABLE "%s" '
+                cursor.execute('CREATE TABLE "%s" '
                     '(id INTEGER PRIMARY KEY AUTOINCREMENT)'
                     % self.table_name)
             else:
-                self.cursor.execute('CREATE TABLE "%s" '
+                cursor.execute('CREATE TABLE "%s" '
                     '(__id INTEGER PRIMARY KEY AUTOINCREMENT, '
                     'id INTEGER)' % self.table_name)
 
         self._update_definitions()
 
     @staticmethod
-    def table_exist(cursor, table_name):
+    def table_exist(table_name):
+        cursor = Transaction().connection.cursor()
         cursor.execute("SELECT sql FROM sqlite_master "
             "WHERE type = 'table' AND name = ?",
             (table_name,))
@@ -68,38 +71,38 @@ class TableHandler(TableHandlerInterface):
         return True
 
     @staticmethod
-    def table_rename(cursor, old_name, new_name):
-        if (TableHandler.table_exist(cursor, old_name)
-                and not TableHandler.table_exist(cursor, new_name)):
+    def table_rename(old_name, new_name):
+        cursor = Transaction().connection.cursor()
+        if (TableHandler.table_exist(old_name)
+                and not TableHandler.table_exist(new_name)):
             cursor.execute('ALTER TABLE "%s" RENAME TO "%s"'
                 % (old_name, new_name))
         # Rename history table
         old_history = old_name + "__history"
         new_history = new_name + "__history"
-        if (TableHandler.table_exist(cursor, old_history)
-                and not TableHandler.table_exist(cursor, new_history)):
+        if (TableHandler.table_exist(old_history)
+                and not TableHandler.table_exist(new_history)):
             cursor.execute('ALTER TABLE "%s" RENAME TO "%s"'
                 % (old_history, new_history))
 
     @staticmethod
-    def sequence_exist(cursor, sequence_name):
+    def sequence_exist(sequence_name):
         return True
 
     @staticmethod
-    def sequence_rename(cursor, old_name, new_name):
+    def sequence_rename(old_name, new_name):
         pass
 
     def column_exist(self, column_name):
         return column_name in self._columns
 
     def column_rename(self, old_name, new_name, exception=False):
+        cursor = Transaction().connection.cursor()
         if self.column_exist(old_name) and \
                 not self.column_exist(new_name):
             temp_table = '_temp_%s' % self.table_name
-            TableHandler.table_rename(self.cursor, self.table_name,
-                temp_table)
-            new_table = TableHandler(self.cursor, self._model,
-                history=self.history)
+            TableHandler.table_rename(self.table_name, temp_table)
+            new_table = TableHandler(self._model, history=self.history)
             for column, (notnull, hasdef, size, typname) \
                     in self._columns.iteritems():
                 if column == old_name:
@@ -109,12 +112,12 @@ class TableHandler(TableHandlerInterface):
             new_columns = new_table._columns.keys()
             old_columns = [x if x != old_name else new_name
                 for x in new_columns]
-            self.cursor.execute(('INSERT INTO "%s" (' +
+            cursor.execute(('INSERT INTO "%s" (' +
                     ','.join('"%s"' % x for x in new_columns) +
                     ') SELECT ' +
                     ','.join('"%s"' % x for x in old_columns) + ' ' +
                     'FROM "%s"') % (self.table_name, temp_table))
-            self.cursor.execute('DROP TABLE "%s"' % temp_table)
+            cursor.execute('DROP TABLE "%s"' % temp_table)
             self._update_definitions()
         elif exception and self.column_exist(new_name):
             raise Exception('Unable to rename column %s.%s to %s.%s: '
@@ -122,62 +125,56 @@ class TableHandler(TableHandlerInterface):
                 % (self.table_name, old_name, self.table_name, new_name,
                     self.table_name, new_name))
 
-    def _update_definitions(self):
-
+    def _update_definitions(self, columns=None, indexes=None):
+        if columns is None and indexes is None:
+            columns = indexes = True
+        cursor = Transaction().connection.cursor()
         # Fetch columns definitions from the table
-        self.cursor.execute('PRAGMA table_info("' + self.table_name + '")')
-        self._columns = {}
-        for _, column, type_, notnull, hasdef, _ in self.cursor.fetchall():
-            column = re.sub(r'^\"|\"$', '', column)
-            match = re.match(r'(\w+)(\((.*?)\))?', type_)
-            if match:
-                typname = match.group(1).upper()
-                size = match.group(3) and int(match.group(3)) or 0
-            else:
-                typname = type_.upper()
-                size = -1
-            self._columns[column] = {
-                'notnull': notnull,
-                'hasdef': hasdef,
-                'size': size,
-                'typname': typname,
-            }
+        if columns:
+            cursor.execute('PRAGMA table_info("' + self.table_name + '")')
+            self._columns = {}
+            for _, column, type_, notnull, hasdef, _ in cursor.fetchall():
+                column = re.sub(r'^\"|\"$', '', column)
+                match = re.match(r'(\w+)(\((.*?)\))?', type_)
+                if match:
+                    typname = match.group(1).upper()
+                    size = match.group(3) and int(match.group(3)) or 0
+                else:
+                    typname = type_.upper()
+                    size = -1
+                self._columns[column] = {
+                    'notnull': notnull,
+                    'hasdef': hasdef,
+                    'size': size,
+                    'typname': typname,
+                }
 
         # Fetch indexes defined for the table
-        try:
-            self.cursor.execute('PRAGMA index_list("' + self.table_name + '")')
-        except IndexError:  # There is sometimes IndexError
-            self.cursor.execute('PRAGMA index_list("' + self.table_name + '")')
-        self._indexes = [l[1] for l in self.cursor.fetchall()]
+        if indexes:
+            try:
+                cursor.execute('PRAGMA index_list("' + self.table_name + '")')
+            except IndexError:  # There is sometimes IndexError
+                cursor.execute('PRAGMA index_list("' + self.table_name + '")')
+            self._indexes = [l[1] for l in cursor.fetchall()]
 
-        # Keep track of which module created each field
-        self._field2module = {}
-        if self.object_name is not None:
-            self.cursor.execute('SELECT f.name, f.module '
-                'FROM ir_model_field f '
-                'JOIN ir_model m on (f.model=m.id) '
-                'WHERE m.model = ?',
-                (self.object_name,))
-            for line in self.cursor.fetchall():
-                self._field2module[line[0]] = line[1]
+    @property
+    def _field2module(self):
+        cursor = Transaction().connection.cursor()
+        cursor.execute('SELECT f.name, f.module '
+            'FROM ir_model_field f '
+            'JOIN ir_model m on (f.model=m.id) '
+            'WHERE m.model = ?',
+            (self.object_name,))
+        return dict(cursor)
 
     def alter_size(self, column_name, column_type):
-        logger.warning(
-            'Unable to alter size of column %s '
-            'of table %s!',
-            column_name, self.table_name)
+        warnings.warn('Unable to alter size of column with SQLite backend')
 
     def alter_type(self, column_name, column_type):
-        logger.warning(
-            'Unable to alter type of column %s '
-            'of table %s!',
-            column_name, self.table_name)
+        warnings.warn('Unable to alter type of column with SQLite backend')
 
     def db_default(self, column_name, value):
-        logger.warning(
-            'Unable to set default on column %s '
-            'of table %s!',
-            column_name, self.table_name)
+        warnings.warn('Unable to set default on column with SQLite backend')
 
     def add_raw_column(self, column_name, column_type, column_format,
             default_fun=None, field_size=None, migrate=True, string=''):
@@ -221,57 +218,55 @@ class TableHandler(TableHandlerInterface):
                         field_size)
             return
 
+        cursor = Transaction().connection.cursor()
         column_type = column_type[1]
         default = ''
-        self.cursor.execute(('ALTER TABLE "%s" ADD COLUMN "%s" %s' + default) %
+        cursor.execute(('ALTER TABLE "%s" ADD COLUMN "%s" %s' + default) %
                        (self.table_name, column_name, column_type))
 
         if column_format:
             # check if table is non-empty:
-            self.cursor.execute('SELECT 1 FROM "%s" limit 1' % self.table_name)
-            if self.cursor.fetchone():
+            cursor.execute('SELECT 1 FROM "%s" limit 1' % self.table_name)
+            if cursor.fetchone():
                 # Populate column with default values:
                 default = None
                 if default_fun is not None:
                     default = default_fun()
-                self.cursor.execute('UPDATE "' + self.table_name + '" '
+                cursor.execute('UPDATE "' + self.table_name + '" '
                     'SET "' + column_name + '" = ?',
                     (column_format(default),))
 
-        self._update_definitions()
+        self._update_definitions(columns=True)
 
     def add_fk(self, column_name, reference, on_delete=None):
-        logger.warning(
-            'Unable to add foreign key on table %s!',
-            self.table_name)
+        warnings.warn('Unable to add foreign key with SQLite backend')
 
     def drop_fk(self, column_name, table=None):
-        logger.warning(
-            'Unable to drop foreign key on table %s!',
-            self.table_name)
+        warnings.warn('Unable to drop foreign key with SQLite backend')
 
     def index_action(self, column_name, action='add', table=None):
         if isinstance(column_name, basestring):
             column_name = [column_name]
         index_name = self.table_name + "_" + '_'.join(column_name) + "_index"
 
+        cursor = Transaction().connection.cursor()
         if action == 'add':
             if index_name in self._indexes:
                 return
-            self.cursor.execute('CREATE INDEX "' + index_name + '" '
+            cursor.execute('CREATE INDEX "' + index_name + '" '
                 'ON "' + self.table_name + '" ( ' +
                 ','.join('"' + x + '"' for x in column_name) +
                 ')')
-            self._update_definitions()
+            self._update_definitions(indexes=True)
         elif action == 'remove':
             if len(column_name) == 1:
-                if self._field2module.get(column_name[0], self.module_name) \
-                        != self.module_name:
+                if self._field2module.get(column_name[0],
+                        self.module_name) != self.module_name:
                     return
 
             if index_name in self._indexes:
-                self.cursor.execute('DROP INDEX "%s" ' % (index_name,))
-                self._update_definitions()
+                cursor.execute('DROP INDEX "%s" ' % (index_name,))
+                self._update_definitions(indexes=True)
         else:
             raise Exception('Index action not supported!')
 
@@ -280,35 +275,24 @@ class TableHandler(TableHandlerInterface):
             return
 
         if action == 'add':
-            logger.warning(
-                'Unable to set not null on column %s '
-                'of table %s!',
-                column_name, self.table_name)
+            warnings.warn('Unable to set not null with SQLite backend')
         elif action == 'remove':
-            logger.warning(
-                'Unable to remove not null on column %s '
-                'of table %s!',
-                column_name, self.table_name)
+            warnings.warn('Unable to remove not null with SQLite backend')
         else:
             raise Exception('Not null action not supported!')
 
     def add_constraint(self, ident, constraint, exception=False):
-        logger.warning(
-            'Unable to add constraint on table %s!',
-            self.table_name)
+        warnings.warn('Unable to add constraint with SQLite backend')
 
     def drop_constraint(self, ident, exception=False, table=None):
-        logger.warning(
-            'Unable to drop constraint on table %s!',
-            self.table_name)
+        warnings.warn('Unable to drop constraint with SQLite backend')
 
     def drop_column(self, column_name, exception=False):
-        logger.warning(
-            'Unable to drop "%s" column on table %s!',
-            column_name, self.table_name)
+        warnings.warn('Unable to drop column with SQLite backend')
 
     @staticmethod
-    def drop_table(cursor, model, table, cascade=False):
+    def drop_table(model, table, cascade=False):
+        cursor = Transaction().connection.cursor()
         cursor.execute('DELETE from ir_model_data where '
             'model = \'%s\'' % model)
 
