@@ -241,23 +241,39 @@ class Many2Many(Field):
         value = tuple(instance(x) for x in (value or []))
         super(Many2Many, self).__set__(inst, value)
 
-    def convert_domain_child(self, domain, tables):
+    def convert_domain_tree(self, domain, tables):
         Target = self.get_target()
         table, _ = tables[None]
         name, operator, ids = domain
-        ids = list(ids)  # Ensure it is a list for concatenation
+        ids = set(ids)  # Ensure it is a set for concatenation
 
         def get_child(ids):
             if not ids:
-                return []
+                return set()
             children = Target.search([
                     (name, 'in', ids),
                     (name, '!=', None),
                     ], order=[])
-            child_ids = get_child([c.id for c in children])
-            return ids + child_ids
-        expression = table.id.in_(ids + get_child(ids))
-        if operator == 'not child_of':
+            child_ids = get_child(set(c.id for c in children))
+            return ids | child_ids
+
+        def get_parent(ids):
+            if not ids:
+                return set()
+            parent_ids = set()
+            for parent in Target.browse(ids):
+                parent_ids.update(p.id for p in getattr(parent, name))
+            return ids | get_parent(parent_ids)
+
+        if operator.endswith('child_of'):
+            ids = list(get_child(ids))
+        else:
+            ids = list(get_parent(ids))
+        if not ids:
+            expression = Literal(False)
+        else:
+            expression = table.id.in_(ids)
+        if operator.startswith('not'):
             return ~expression
         return expression
 
@@ -270,6 +286,7 @@ class Many2Many(Field):
         transaction = Transaction()
         table, _ = tables[None]
         name, operator, value = domain[:3]
+        assert operator not in {'where', 'not where'} or '.' not in name
 
         if Relation._history and transaction.context.get('_datetime'):
             relation = Relation.__table_history__()
@@ -290,10 +307,15 @@ class Many2Many(Field):
 
         target = getattr(Relation, self.target).sql_column(relation)
         if '.' not in name:
-            if operator in ('child_of', 'not child_of'):
+            if operator.endswith('child_of') or operator.endswith('parent_of'):
                 if Target != Model:
-                    query = Target.search([(domain[3], 'child_of', value)],
-                        order=[], query=True)
+                    if operator.endswith('child_of'):
+                        target_operator = 'child_of'
+                    else:
+                        target_operator = 'parent_of'
+                    query = Target.search([
+                            (domain[3], target_operator, value),
+                            ], order=[], query=True)
                     where = (target.in_(query) & (origin != Null))
                     if history_where:
                         where &= history_where
@@ -301,7 +323,7 @@ class Many2Many(Field):
                         where &= origin_where
                     query = relation.select(origin, where=where)
                     expression = table.id.in_(query)
-                    if operator == 'not child_of':
+                    if operator.startswith('not'):
                         return ~expression
                     return expression
                 if isinstance(value, basestring):
@@ -313,12 +335,12 @@ class Many2Many(Field):
                 else:
                     ids = value
                 if not ids:
-                    expression = table.id.in_([None])
-                    if operator == 'not child_of':
+                    expression = Literal(False)
+                    if operator.startswith('not'):
                         return ~expression
                     return expression
                 else:
-                    return self.convert_domain_child(
+                    return self.convert_domain_tree(
                         (name, operator, ids), tables)
 
             if value is None:
@@ -340,11 +362,18 @@ class Many2Many(Field):
         else:
             _, target_name = name.split('.', 1)
 
-        relation_domain = [('%s.%s' % (self.target, target_name),)
-            + tuple(domain[1:])]
-        if origin_field._type == 'reference':
-            relation_domain.append(
-                (self.origin, 'like', Model.__name__ + ',%'))
+        if operator not in {'where', 'not where'}:
+            relation_domain = [('%s.%s' % (self.target, target_name),)
+                + tuple(domain[1:])]
+            if origin_field._type == 'reference':
+                relation_domain.append(
+                    (self.origin, 'like', Model.__name__ + ',%'))
+        else:
+            relation_domain = []
+            for clause in value:
+                relation_domain.append(
+                        ('%s.%s' % (self.target, clause[0]),)
+                        + tuple(clause[1:]))
         rule_domain = Rule.domain_get(Relation.__name__, mode='read')
         if rule_domain:
             relation_domain = [relation_domain, rule_domain]
@@ -355,4 +384,8 @@ class Many2Many(Field):
             relation_domain, tables=relation_tables)
         query_table = convert_from(None, relation_tables)
         query = query_table.select(origin, where=expression)
-        return table.id.in_(query)
+        expression = table.id.in_(query)
+
+        if operator == 'not where':
+            expression = ~expression
+        return expression
